//
// Written by Boguslaw Cyganek
// Copyrights (C) 2006
//


// Circle.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include <iostream.h>



/*class MyData
{
	private:

		// Class inherent variables

		double fData;			


	protected:

		MyData() {}

	public:

		MyData( double d ) : fData( d ) {}
};


*/
///////////////////////////////////////////////////////////
// This class models a circle.
///////////////////////////////////////////////////////////
class TCircle
{
	private:

		// Class inherent variables

		double fRadius;				// contains a radius of a circle
//		MyData fRadius;				// contains a radius of a circle

	protected:

		///////////////////////////////////////////////////////////
		// This function returns the constant PI up to two
		// places after the dot.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			limited approximation of PI (=3.14)
		//		
		// REMARKS:
		//			This function is virtual.
		//		
		virtual double GetPi( void ) const { return 3.14; }		// be careful, do not write 3,14 !!!

	public:

		// ===================================================
		
		TCircle( void );				// class default constructor

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			radius - radius of a circle
		//		
		// OUTPUT:
		//			
		//		
		// REMARKS:
		//			REMEMBER that in the constructor after : we can initialize  
		//				members of a class, we can call the base class
		//				constructor, however
		//				we MUST initialize here all const values
		//				as well as the reference values !!!!!!!!
		//		
		TCircle( double r );			

		TCircle( const TCircle & );						// class copy constructor

		TCircle & operator = ( const TCircle & );		// class assignement operator

		virtual ~TCircle();								// class virtual destructor

		// ===================================================

	public:


		///////////////////////////////////////////////////////////
		// Getter
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			radius of a circle
		//		
		// REMARKS:
		//		
		//		
		double GetRadius( void ) const { return fRadius; }

		///////////////////////////////////////////////////////////
		// Setter
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			radius of a circle
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		void SetRadius( double r ) { fRadius = r; }

	public:

		///////////////////////////////////////////////////////////
		// This function returns area of a circle
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			area of this circle with limited
		//				accuracy
		//		
		// REMARKS:
		//		
		//		
		double GetArea( void ) const;
};


///////////////////////////////////////////////////////////
// Class default constructor 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TCircle::TCircle( void )		// class constructor
	: fRadius( 0 )
{
}

///////////////////////////////////////////////////////////
// Class constructor 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			radius - radius of a circle
//		
// OUTPUT:
//			
//		
// REMARKS:
//			REMEMBER that in the constructor after : we can initialize  
//				members of a class, we can call the base class
//				constructor, however
//				we MUST initialize here all const values
//				as well as the reference values !!!!!!!!
//		
TCircle::TCircle( double radius )		
	: fRadius( radius )					
{
}	// this is empty body of this function

///////////////////////////////////////////////////////////
// Class destructor 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TCircle::~TCircle()		// class destructor
{
#if _DEBUG
	cout << "The ~TCircle() was just called..." << endl;
#endif
}




///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TCircle::TCircle( const TCircle & ref )	// class copy constructor
	: fRadius( ref.fRadius )
{
}


///////////////////////////////////////////////////////////
// This is an assignement operator 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			ref - reference to the TCircle
//		
// OUTPUT:
//			this
//		
// REMARKS:
//		
//		
TCircle & TCircle::operator = ( const TCircle & ref )	// class assignement operator
{
	if( this != & ref )
	{
		// i.e. we have a different object
	
		fRadius = ref.fRadius;
	}

	return * this;
}


///////////////////////////////////////////////////////////
// This function returns area of a circle
///////////////////////////////////////////////////////////
//		
// INPUT:
//			none
//		
// OUTPUT:
//			area of this circle with limited
//				accuracy
//		
// REMARKS:
//		
//		
double TCircle::GetArea( void ) const
{
	return fRadius * fRadius * GetPi();
}


// ==============================================================================================================

///////////////////////////////////////////////////////////
// This class models a more precise circle.
///////////////////////////////////////////////////////////
class TPreciseCircle : public TCircle
{

	protected:

		///////////////////////////////////////////////////////////
		// This function returns the constant PI up to five
		// places after the dot.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			limited approximation of PI (=3.14151)
		//		
		// REMARKS:
		//			This function is virtual.
		//		
		virtual double GetPi( void ) const { return 3.14151; }		// be careful, do not write 3,14 !!!

	public:

		// ===================================================

		TPreciseCircle( void ) {}									// class default constructor

		// The fRadius member from the base class is declared private.
		// Therefore we cannot initialize it directly, even in this derived class.
		// However, we can call the base class constructor TCircle and
		// it will do the job for us.
		TPreciseCircle( double r ) : TCircle( r )					
		{
		}		
		
		// Here after : we also call the base class constructor. 
		// This time it is a copy constructor, however.
		TPreciseCircle( const TPreciseCircle & c ) : TCircle( c ) {}		// class copy constructor

		TPreciseCircle & operator = ( const TPreciseCircle & c )	// class assignement operator
		{
			TCircle::operator = ( c );								// let's delegate an action to the base class
			return * this;
		}

		virtual ~TPreciseCircle()									// class virtual destructor
		{
			#if _DEBUG
				cout << "The ~TPreciseCircle() was just called..." << endl;
			#endif
		}				

		// ===================================================


};


// ==============================================================================================================



int main(int argc, char* argv[])
{
	TCircle c1, c2( 13.13 );

	TPreciseCircle pc1, pc2( 13.13 );

	// Let us check assignements
	//pc1 = c1;			// can we assign a base to its derivative? Why not?
	c2 = pc2;
	pc1 = pc1;
	pc1 = pc2;

	// Let us check GetArea(). Go inside and check what GetPi() is called?
	cout << "Area of the circle is: " << c2.GetArea() << endl;
	cout << "Area of the precise circle is: " << pc2.GetArea() << endl;

	cout << endl << endl;

	// More advanced stuff goes here...
	// Let's create two objects on the heap, however the pointers
	// to them are pointers to the base class !

	TCircle * circle_1 = new TCircle( 10.0 );
	TCircle * prec_circle_2 = new TPreciseCircle( 20.0 );

	
	cout << "Area of the circle is: " << circle_1->GetArea() << endl;
	cout << "Area of the circle is: " << prec_circle_2->GetArea() << endl;


	// Do not forget to get rid of the allocated objects
	delete prec_circle_2;		// but what destructor will be called here since prec_circle_2 is of type: TCircle * 
	delete circle_1;			// What happens if we omitted the key word "virtual" before the destructors?




	// This is to stop the program for awhile...
	int dummy;
	cin >> dummy;

	return 0;
}

// ==============================================================================================================





